from .exception import XmlSchemaException
from .config import Config, TypeGenerator
from .xpath import NormalizedXPath

from fences.core.node import Leaf, Decision, NoOpLeaf, NoOpDecision, Node, Reference
from fences.core.random import generate_random_number, generate_random_string, StringProperties

from xml.etree import ElementTree
from typing import Set, List, Optional, Dict


def default_config() -> Config:
    return Config(
        tag_handlers={
            'all': parse_sequence,  # TODO: special handling
            'element': parse_element,
            'sequence': parse_sequence,
            'choice': parse_choice,
            'simpleType': parse_type,
            'complexType': parse_type,
            'simpleContent': parse_content,
            'complexContent': parse_content,
            'attribute': parse_attribute,
            'annotation': parse_annotation,
            'extension': parse_extension,
            'restriction': parse_restriction,
            'any': parse_any,
        },
        type_generators={
            'xs:string': TypeGenerator(
                valid=lambda *_: [
                    'foo', generate_random_string(StringProperties(1))],
                invalid=lambda *_: []
            ),
            'xs:dateTime': TypeGenerator(
                valid=lambda *_: ['2001-10-26T21:32:52'],
                invalid=lambda *_: ["foo"]
            ),
            'xs:positiveInteger': TypeGenerator(
                valid=lambda *_: [generate_random_number(min_value=0)],
                invalid=lambda *_: ["-10", "foo"]
            ),
            'xs:integer': TypeGenerator(
                valid=lambda *_: [generate_random_number(min_value=0)],
                invalid=lambda *_: ["xyz"]
            ),
            'xs:boolean': TypeGenerator(
                valid=lambda *_: ["true", "false", "0", "1"],
                invalid=lambda *_: ["foo"]
            ),
            'xs:unsignedInt': TypeGenerator(
                valid=lambda *_: [generate_random_number(min_value=0)],
                invalid=lambda *_: ["-10", "bar"]
            ),
            'xs:unsignedShort': TypeGenerator(
                valid=lambda *_: [generate_random_number(min_value=0)],
                invalid=lambda *_: ["-10", "bar"]
            ),
            'xs:unsignedByte': TypeGenerator(
                valid=lambda *_: [generate_random_number(min_value=0, max_value=255)],
                invalid=lambda *_: ["-10", "bar"]
            ),
            'xs:int': TypeGenerator(
                valid=lambda *_: [generate_random_number(min_value=0)],
                invalid=lambda *_: ["bar"]
            ),
            'xs:double': TypeGenerator(
                valid=lambda *_: [generate_random_number(min_value=0)],
                invalid=lambda *_: ["bar"]
            ),
            'xs:decimal': TypeGenerator(
                valid=lambda *_: [generate_random_number()],
                invalid=lambda *_: ["bar"]
            ),
        },
        restriction_handlers={
            'pattern': parse_string_restriction,
            'minLength': parse_string_restriction,
            'maxLength': parse_string_restriction,
        }
    )


class Binding:
    def __init__(self, element: ElementTree.Element, attr: Optional[str]) -> None:
        self.element = element
        self.attr = attr

    def set_value(self, value: any):
        if self.attr:
            self.element.attrib[self.attr] = value
        else:
            self.element.text = value


class StartNode(Decision):

    def apply(self, data: any) -> Binding:
        return Binding(ElementTree.Element('dummy'), None)

    def description(self) -> str:
        return "Start"


class FetchOutput(Leaf):
    def __init__(self, namespace: Optional[str]) -> None:
        super().__init__(None, True)
        self.namespace = namespace

    def apply(self, data: Binding) -> ElementTree.ElementTree:
        root = next(iter(data.element))
        if self.namespace:
            root.attrib['xmlns'] = self.namespace
        return ElementTree.ElementTree(root)

    def description(self) -> str:
        return "Fetch output"


class StartAttribute(Decision):
    def __init__(self, id: str, all_transitions: bool, attr: str) -> None:
        super().__init__(id, all_transitions)
        self.attr = attr

    def apply(self, data: Binding) -> Binding:
        return Binding(data.element, self.attr)

    def description(self) -> Optional[str]:
        return f"Start attribute '{self.attr}'"


class StartNewElement(Decision):

    def __init__(self, id: str, all_transitions: bool, tag: str) -> None:
        super().__init__(id, all_transitions)
        self.tag = tag

    def apply(self, data: Binding) -> ElementTree.Element:
        element = ElementTree.Element(self.tag)
        data.element.append(element)
        return Binding(element, None)

    def description(self) -> str:
        return f"Start <{self.tag}..."


class SetValueLeaf(Leaf):
    def __init__(self, id: str, is_valid: bool, value: any) -> None:
        super().__init__(id, is_valid)
        self.value = value

    def apply(self, data: Binding) -> any:
        data.set_value(self.value)
        return data

    def description(self) -> Optional[str]:
        return f"= '{self.value}'"


def _get_tag(el: ElementTree.Element) -> str:
    _, _, tag_wo_namespace = el.tag.rpartition('}')
    return tag_wo_namespace


def _lookup(attrib: Dict[str, str], key: str, parsed_attributes: Set[str]) -> str:
    parsed_attributes.remove(key)
    return attrib[key]


def parse_enumeration_restriction(element: ElementTree.Element, path) -> Node:
    if any(_get_tag(child) != 'enumeration' for child in element):
        raise XmlSchemaException("Only enumeration allowed", path)

    root = NoOpDecision(str(path), False)
    for child in element:
        value = child.attrib['value']
        root.add_transition(SetValueLeaf(None, True, value))
    return root


def parse_string_restriction(base: str, props: Dict[str, str], config: Config, parsed_props: Set[str], path: NormalizedXPath) -> Node:
    if base not in ['xs:string', 'xs:token']:
        raise XmlSchemaException(
            f"String restrictions must be based on string class, is {base}",
            path)

    s = StringProperties()
    if 'pattern' in props:
        s.pattern = _lookup(props, 'pattern', parsed_props)
    if 'minLength' in props:
        s.min_length = int(_lookup(props, 'minLength', parsed_props))
    if 'maxLength' in props:
        s.max_length = int(_lookup(props, 'maxLength', parsed_props))

    return SetValueLeaf(str(path), True, generate_random_string(s))


def parse_any(element: ElementTree.Element, config: Config, parsed_attributes: Set[str], path: NormalizedXPath) -> Node:
    if 'processContents' in element.attrib:
        process_contents = _lookup(
            element.attrib, 'processContents', parsed_attributes)
    # TODO: use processContents
    return NoOpLeaf(str(path), True)


def parse_restriction(element: ElementTree.Element, config: Config, parsed_attributes: Set[str], path: NormalizedXPath) -> Node:
    base = _lookup(element.attrib, 'base', parsed_attributes)

    # Normally, we would use base to check if we really restrict the base type
    # However, Fences is not meant to be a schema checker, so we discard it here

    if len(element) == 0:
        return resolve_type(base, element, config, path)

    # Special handling for enumerations
    if _get_tag(next(iter(element))) == 'enumeration':
        return parse_enumeration_restriction(element, path)

    handler = None
    for child in element:
        tag = _get_tag(child)
        if tag in config.restriction_handlers:
            handler = config.restriction_handlers[tag]
            break
    if handler is None:
        child_tags = [_get_tag(i) for i in element]
        raise XmlSchemaException(
            f"No restriction handler found, child tags are {child_tags}", path)

    props = {}
    for child in element:
        if len(child.attrib) != 1 or 'value' not in child.attrib:
            raise XmlSchemaException(
                "Restriction children must have exactly one attribute 'value'", path)
        key = _get_tag(child)
        value = child.attrib['value']
        if key in props:
            raise XmlSchemaException(
                f"Duplicate restriction value '{key}'", path)
        props[key] = value

    parsed_props = set(props.keys())
    node = handler(base, props, config, parsed_props, path)
    if parsed_props:
        raise XmlSchemaException(
            f"Unhandled restriction properties {parsed_props}", path)

    # TODO: in case of contentContent/restriction we have to parse the children as well!
    return node


def parse_annotation(element: ElementTree.Element, config: Config, parsed_attributes: Set[str], path: NormalizedXPath) -> Node:
    return NoOpLeaf(str(path), True)


def parse_content(element: ElementTree.Element, config: Config, parsed_attributes: Set[str], path: NormalizedXPath) -> Node:
    root = NoOpDecision(str(path), True)
    for subpath, child in path.enumerate(element):
        root.add_transition(
            parse_xml_element(child, config, subpath)
        )
    return root


def parse_extension(element: ElementTree.Element, config: Config, parsed_attributes: Set[str], path: NormalizedXPath) -> Node:
    root = NoOpDecision(str(path), True)
    base = _lookup(element.attrib, 'base', parsed_attributes)

    root.add_transition(
        resolve_type(base, element, config, path)
    )

    for subpath, child in path.enumerate(element):
        root.add_transition(
            parse_xml_element(child, config, subpath)
        )

    return root


def _parse_occurs(element: ElementTree.Element, path: NormalizedXPath, parsed_attributes: Set[str]):
    if 'minOccurs' in element.attrib:
        s = _lookup(element.attrib, 'minOccurs', parsed_attributes)
        try:
            min_occurs = int(s)
        except ValueError as e:
            raise XmlSchemaException(
                f"Invalid value for 'minOccurs': {e}", path)
    else:
        min_occurs = 1

    if 'maxOccurs' in element.attrib:
        s = _lookup(element.attrib, 'maxOccurs', parsed_attributes)
        if s == 'unbounded':
            max_occurs = None
        else:
            try:
                max_occurs = int(s)
            except ValueError as e:
                raise XmlSchemaException(
                    f"Invalid value for 'maxOccurs': {e}", path)
    else:
        max_occurs = 1

    return min_occurs, max_occurs


def _repeat(child: Node, min_occurs: int, max_occurs: Optional[int]) -> Node:

    root = NoOpDecision(None, False)
    root.add_transition(NoOpLeaf(None, is_valid=min_occurs == 0))

    # Check parameters
    if max_occurs is None:
        max_occurs = min_occurs + 1
    if min_occurs > max_occurs:
        raise XmlSchemaException(
            f"minOccurs = {min_occurs} > maxOccurs = {min_occurs}", None)

    # valid: min_occurs
    if min_occurs > 0:
        subroot = NoOpDecision(None, True)
        for _ in range(min_occurs):
            subroot.add_transition(child)
        root.add_transition(subroot)

    # invalid: min_occurs - 1
    if min_occurs > 1:
        subroot = NoOpDecision(None, True)
        for _ in range(min_occurs - 1):
            subroot.add_transition(child, True)
        root.add_transition(subroot, False)

    if max_occurs != min_occurs:
        # Valid: max_occurs
        subroot = NoOpDecision(None, True)
        for _ in range(max_occurs):
            subroot.add_transition(child)
        root.add_transition(subroot)

    return root


def parse_sequence(element: ElementTree.Element, config: Config, parsed_attributes: Set[str], path: NormalizedXPath) -> Node:
    if 'name' in element.attrib:
        name = _lookup(element.attrib, 'name', parsed_attributes)
    else:
        name = None
    root = NoOpDecision(str(path), True)
    for subpath, i in path.enumerate(element):
        root.add_transition(parse_xml_element(i, config, subpath))
    return root


def parse_choice(element: ElementTree.Element, config: Config, parsed_attributes: Set[str], path: NormalizedXPath) -> Node:
    if 'name' in element.attrib:
        name = _lookup(element.attrib, 'name', parsed_attributes)
    else:
        name = None
    root = NoOpDecision(str(path), False)
    for subpath, i in path.enumerate(element):
        root.add_transition(parse_xml_element(i, config, subpath))
    return root


def parse_attribute(element: ElementTree.Element, config: Config, parsed_attributes: Set[str], path: NormalizedXPath) -> Node:
    if 'ref' in element.attrib:
        raise XmlSchemaException("references currently not supported", path)

    name = _lookup(element.attrib, 'name', parsed_attributes)
    required = False  # Attributes are optional by default
    if 'use' in element.attrib and _lookup(element.attrib, 'use', parsed_attributes) == "required":
        required = True
    # TODO: can we combine required with fixed?

    if 'default' in element.attrib:
        if required:
            raise XmlSchemaException(
                "Cannot set a default value for a required attribute", path)
        _lookup(element.attrib, 'default', parsed_attributes)

    super_root = NoOpDecision(str(path), False)
    super_root.add_transition(NoOpLeaf(None, is_valid=not required))

    root = StartAttribute(None, False, name)
    super_root.add_transition(root)

    if 'fixed' in element.attrib:
        if 'type' in element.attrib:
            # Discard the type
            _lookup(element.attrib, 'type', parsed_attributes)
        if 'default' in element.attrib:
            raise XmlSchemaException(
                f"Cannot set default value for fixed attribute", path)
        fixed_value = _lookup(element.attrib, 'fixed', parsed_attributes)
        root.add_transition(SetValueLeaf(None, True, fixed_value))
        root.add_transition(SetValueLeaf(
            None, False, fixed_value + '_INVALID'))
    else:
        if 'type' in element.attrib:
            if len(element) != 0:
                raise XmlSchemaException(
                    f"Attribute with type cannot have children", path)
            type = _lookup(element.attrib, 'type', parsed_attributes)
            root.add_transition(resolve_type(type, element, config, path))
        else:
            for subpath, i in path.enumerate(element):
                root.add_transition(parse_xml_element(i, config, subpath))

    return super_root


def parse_type(element: ElementTree.Element, config: Config, parsed_attributes: Set[str], path: NormalizedXPath) -> Node:
    if 'name' in element.attrib:
        name = _lookup(element.attrib, 'name', parsed_attributes)
    else:
        name = None
    root = NoOpDecision(name, True)

    for subpath, child in path.enumerate(element):
        child_node = parse_xml_element(child, config, subpath)
        root.add_transition(child_node)

    return root


def parse_element(element: ElementTree.Element, config: Config, parsed_attributes: Set[str], path: NormalizedXPath) -> Node:
    name = _lookup(element.attrib, 'name', parsed_attributes)
    if 'type' in element.attrib:
        type = _lookup(element.attrib, 'type', parsed_attributes)
        root = StartNewElement(str(path), False, name)
        root.add_transition(
            resolve_type(type, element, config, path)
        )
        return root
    else:
        if len(element) != 1:
            raise XmlSchemaException(f"Expected exactly one child", path)
        root = StartNewElement(str(path), False, name)
        for subpath, child in path.enumerate(element):
            child_node = parse_xml_element(child, config, subpath)
        root.add_transition(child_node)
        return root


def _remove_if_available(s: set, key: str):
    if key in s:
        s.remove(key)


def parse_xml_element(element: ElementTree.Element, config: Config, path: NormalizedXPath) -> Node:
    tag = _get_tag(element)
    try:
        handler = config.tag_handlers[tag]
    except KeyError:
        raise XmlSchemaException(f"Unknown tag '{tag}'", path)
    parsed_attributes = set(element.attrib.keys())
    node = handler(element, config, parsed_attributes, path)

    if 'minOccurs' in element.attrib or 'maxOccurs' in element.attrib:
        min_occurs, max_occurs = _parse_occurs(element, path, parsed_attributes)
        node = _repeat(node, min_occurs, max_occurs)

    if parsed_attributes:
        raise XmlSchemaException(
            f"Unknown attributes '{parsed_attributes}' in <{tag}>", path)

    return node


def resolve_type(type: str, element: ElementTree.Element, config: Config, path: NormalizedXPath) -> Node:
    if type in config.type_generators:
        handler = config.type_generators[type]
        root = NoOpDecision(None, False)
        for value in handler.valid(element, path):
            value_node = SetValueLeaf(None, True, str(value))
            root.add_transition(value_node)
        for value in handler.invalid(element, path):
            value_node = SetValueLeaf(None, False, str(value))
            root.add_transition(value_node)
        return root
    else:
        # We assume the type is a user defined simple or complex type
        return Reference(None, type)


def parse(schema: ElementTree.Element, config: Optional[dict] = None) -> Node:
    actual_config = default_config()
    if config is not None:
        actual_config.merge(config)
    path = NormalizedXPath([('schema', 0)])
    tag = _get_tag(schema)
    if tag != 'schema':
        raise XmlSchemaException(f"Expected tag 'schema', got '{tag}'", path)
    target_namespace = schema.get('targetNamespace', None)

    # The actual schemas
    element_nodes: List[Node] = []
    other_nodes: List[Node] = []
    for subpath, i in path.enumerate(schema):
        node = parse_xml_element(i, actual_config, subpath)
        if _get_tag(i) == 'element':
            element_nodes.append(node)
        else:
            other_nodes.append(node)

    if len(element_nodes) == 0:
        raise XmlSchemaException("No root element found", None)
    root = element_nodes.pop(0)
    root = root.resolve(element_nodes + other_nodes)
    root.optimize()

    super_root = StartNode(str(path), True)
    super_root.add_transition(root)
    super_root.add_transition(FetchOutput(target_namespace))
    return super_root



from .json_schema.parse import parse as parse_json_schema
from .regex.parse import parse as parse_regex
from .xml_schema.parse import parse as parse_xml_schema
from .grammar.convert import convert as parse_grammar



from dataclasses import dataclass, field
from typing import Dict, Callable, Set, Generator
from .xpath import NormalizedXPath

from fences.core.node import Node
from fences.core.exception import ConfigException

from xml.etree.ElementTree import Element

TagCallback = Callable[[Element,
                        "Config", Set[str], NormalizedXPath], Node]


def empty_iter(*args):
    return
    yield


@dataclass
class TypeGenerator:
    valid: Callable[[Element, NormalizedXPath], Generator[any, None, None]]
    invalid: Callable[[Element, NormalizedXPath], Generator[any, None, None]]


@dataclass
class Config:
    tag_handlers: Dict[str, TagCallback] = field(default_factory=dict)
    type_generators: Dict[str, TypeGenerator] = field(default_factory=dict)
    restriction_handlers: Dict[str, TagCallback] = field(default_factory=dict)

    def merge(self, data: dict):

        if not isinstance(data, dict):
            raise ConfigException("config must be a dict")

        # parse type_generators
        type_generators = data.get('type_generators', {})
        if not isinstance(type_generators, dict):
            raise ConfigException("type_generators must be a dict")
        for type_name, entry in type_generators.items():
            if not isinstance(entry, dict):
                raise ConfigException(
                    "type_generators entries must be dicts")
            if type_name not in self.type_generators:
                self.type_generators[type_name] = TypeGenerator(
                    empty_iter, empty_iter)
            if 'valid' in entry:
                valid = entry['valid']
                if not isinstance(valid, Callable):
                    raise ConfigException(
                        "type_generators.valid must be callable")
                self.type_generators[type_name].valid = valid
            if 'invalid' in entry:
                invalid = entry['valid']
                if not isinstance(invalid, Callable):
                    raise ConfigException(
                        "type_generators.invalid must be callable")
                self.type_generators[type_name].invalid = invalid
        # TODO parse tag_handlers
        # TODO parse restriction_handlers



import pydot

from .exception import InternalException
from .node import Node, Decision, Leaf, Reference

def _to_graph_node(node: Node):
    attrs = {}
    lines = []
    if node.id:
        lines.append(node.id)
    d = node.description()
    if d:
        lines.append(d)
    if isinstance(node, Decision):
        attrs['shape'] = 'rect'
        if node.all_transitions:
            lines.append("ALL")
    elif isinstance(node, Reference):
        attrs['fillcolor'] = 'yellow'
        attrs['style'] = 'filled'
        lines.append(f"-> {node.reference}")
    elif isinstance(node, Leaf):
        attrs['color'] = 'green' if node.is_valid else 'red'
    else:
        raise InternalException(f"Unknown type '{type(node)}'") # pragma: no cover
    return pydot.Node(str(id(node)), label="\n".join(lines), **attrs)

def render(node: Node) -> pydot.Dot:
    graph = pydot.Dot("my_graph", graph_type="digraph")

    for i in node.items():
        graph.add_node(_to_graph_node(i))
        if isinstance(i, Decision):
            for idx, transition in enumerate(i.outgoing_transitions):
                attrs = {
                    'label': idx
                }
                if transition.is_inverting:
                    attrs['arrowhead'] = 'empty'
                graph.add_edge(pydot.Edge(str(id(i)), str(id(transition.target)), **attrs))

    return graph



from .node import Node, Decision
from .exception import FencesException

class ConsistencyException(FencesException):
    pass

def check_consistency(root: Node):
    """
    Checks, if the node is consistent
    (for debugging purpose)
    """
    # check if all transitions are wired up correctly
    for node in root.items():

        # check incoming transitions
        for transition in node.incoming_transitions:
            idx = transition.outgoing_idx
            if not isinstance(transition.source, Decision):
                raise ConsistencyException(f"Source of incoming transition must be a Decision")
            if idx >= len(transition.source.outgoing_transitions):
                raise ConsistencyException(f"Outgoing index out of range")
            actual = transition.source.outgoing_transitions[idx].target
            if node is not actual:
                raise ConsistencyException(f"Invalid transition at ${node.id}: != {actual.id}")

        # check outgoing transitions
        if isinstance(node, Decision):
            for idx, transition in enumerate(node.outgoing_transitions):
                found = None
                for t in transition.target.incoming_transitions:
                    if t.outgoing_idx == idx and t.source is node:
                        found = True
                        break
                if not found:
                    raise ConsistencyException(f"No corresponding incoming transition")

    # check if all ids are unique
    all_ids = set()
    for node in root.items():
        if node.id is not None:
            if node.id in all_ids:
                raise ConsistencyException(f"Duplicate id '{node.id}'")
            all_ids.add(node.id)

    # check that each node has at least one valid leaf
    # The path generation would fail otherwise
    # TODO



from typing import Dict, List, Optional, Tuple
from .exceptions import CharacterRangeException


class RightHandSide:
    def __add__(self, other: "RightHandSide") -> "Concatenation":
        return Concatenation([self, other])

    def __radd__(self, other: "RightHandSide") -> "Concatenation":
        return Concatenation([other, self])

    def __or__(self, other: "RightHandSide") -> "Alternative":
        return Alternative([self, other])

    def __ror__(self, other: "RightHandSide") -> "Alternative":
        return Alternative([other, self])

    def __mul__(self, other: Tuple[int, Optional[int]]) -> "Repetition":
        assert len(other) == 2
        return Repetition(self, other[0], other[1])

# Support for +


class Concatenation(RightHandSide):

    def __init__(self, elements: List[RightHandSide]) -> None:
        self.elements = elements

    def __add__(self, other: RightHandSide) -> "Concatenation":
        if isinstance(other, Concatenation):
            return Concatenation(self.elements + other.elements)
        else:
            return Concatenation(self.elements + [other])


# Support for * (start, stop)


class Repetition(RightHandSide):
    def __init__(self, element: RightHandSide, start: int, stop: Optional[int]):
        self.element = element
        self.start = start
        self.stop = stop


# Support for |

class Alternative(RightHandSide):
    def __init__(self, elements: List[RightHandSide]):
        self.elements = elements

    def __or__(self, other: RightHandSide) -> "Alternative":
        if isinstance(other, Alternative):
            return Alternative(self.elements + other.elements)
        else:
            return Alternative(self.elements + [other])


# Other

class NonTerminal(RightHandSide):
    def __init__(self, name: str):
        self.name = name


class Terminal(RightHandSide):
    def __init__(self, value: str):
        self.value = value


class CharacterRange(RightHandSide):

    UNICODE_MAX = 0x10FFF

    def __init__(self, start: Optional[str], stop: Optional[str]):
        if start is None:
            self.start = 0
        else:
            if len(start) != 1:
                raise CharacterRangeException(
                    f"start of range must have len 1, has {len(start)}")
            self.start = ord(start)

        if stop is None:
            self.stop = CharacterRange.UNICODE_MAX
        else:
            if len(stop) != 1:
                raise CharacterRangeException(
                    f"stop of range must have len 1, has {len(stop)}")
            self.stop = ord(stop)

        if self.start > self.stop:
            raise CharacterRangeException(
                f"stop must be after start")


Grammar = Dict[NonTerminal, RightHandSide]



from fences.core.node import Decision, Node, Leaf, NoOpDecision, NoOpLeaf
from .exception import RegexException, InternalParserException
from typing import Set, Dict, Union
import string

from .grammar import Lark_StandAlone, Tree, Token, LarkError
_parser = Lark_StandAlone()


class Repetition:
    Infinity = 'inf'

    def __init__(self, min: int, max: Union[int, None]) -> None:
        self.min = min
        self.max = max

    def __str__(self) -> str:
        return f"[{self.min}, {self.max}]"


class CreateInputNode(Decision):
    def __init__(self) -> None:
        super().__init__(None, False)

    def apply(self, data: any) -> any:
        return []

    def description(self) -> str:
        return "Create input"


class FetchOutputNode(Leaf):
    def __init__(self) -> None:
        super().__init__(None, True)

    def apply(self, data: any) -> any:
        return "".join(data)

    def description(self) -> str:
        return "Fetch output"


class AppendCharsLeaf(Leaf):

    def __init__(self, id: str, is_valid: bool, char: str) -> None:
        super().__init__(id, is_valid)
        self.char = char

    def apply(self, data: list) -> any:
        data.append(self.char)
        return data

    def mask(self, s: str) -> str:
        allowed = string.ascii_letters + string.digits
        return "".join([
            (i if i in allowed else f"chr({ord(i)})") for i in s
        ])

    def description(self) -> str:
        return f"Append {self.mask(self.char)}"


def _parse_repetition(tree: Tree[Token]) -> "MatchConverter.Repetition":
    assert len(tree.children) in [1, 2]
    repetition_type = tree.children[0]
    assert isinstance(repetition_type, Tree)
    assert repetition_type.data == 'repetition_type'
    assert len(repetition_type.children) == 1
    qualifier = repetition_type.children[0]
    if isinstance(qualifier, Tree):
        if qualifier.data == 'range':
            assert len(qualifier.children) in [1, 2]
            first = qualifier.children[0]
            assert isinstance(first, Token)
            first_value = int(first.value)

            if len(qualifier.children) == 2:
                second = qualifier.children[1]
                assert isinstance(second, Token)
                second_value = int(second.value)
            else:
                second_value = first_value
            if first_value > second_value:
                raise RegexException(
                    f"Invalid range: {first_value} > {second_value}")
            return Repetition(first_value, second_value)
        raise InternalParserException(
            f"Unknown type {type(qualifier)} ('{qualifier.data}')")
    elif isinstance(qualifier, Token):
        if qualifier.type == 'ZERO_OR_MORE_QUANTIFIER':
            return Repetition(0, None)
        elif qualifier.type == 'ONE_OR_MORE_QUANTIFIER':
            return Repetition(1, None)
        elif qualifier.type == 'ZERO_OR_ONE_QUANTIFIER':
            return Repetition(0, 1)
        raise InternalParserException(
            f"Unknown type {type(qualifier)} ('{qualifier.type}')")
    raise InternalParserException(f"Unknown type {type(qualifier)}")


def _add_repetition(root: Decision, item: Node, times: int):
    assert times >= 0
    subroot = NoOpDecision(None, True)
    for _ in range(times):
        subroot.add_transition(item)
    root.add_transition(subroot)


def _repeat(root: Decision, item: Node, rep: Repetition):
    # Instantiate open ranges
    if rep.max is None:
        rep.max = rep.min + 2
    assert rep.min is not None

    # Special case: do nothing
    # TODO: this invalid for rep.min == 0
    if rep.min == 0 or rep.max == 0:
        root.add_transition(NoOpLeaf(None, True))

    # TODO: invalid: rep.min-1 and rep.max+1
    if rep.min > 0:
        _add_repetition(root, item, rep.min)
    if rep.max != rep.min:
        _add_repetition(root, item, rep.max)


class TreeConverter:

    children: Dict[str, "TreeConverter"] = {}
    all_transitions = False

    def __init__(self, tree: Tree[Token]) -> None:
        assert isinstance(tree, Tree)
        self.tree = tree

    def convert(self, child: Union[Tree, Token]) -> Node:
        if isinstance(child, Tree):
            try:
                converter = self.children[str(child.data)]
            except KeyError as e:
                raise InternalParserException(
                    f"Unexpected token {e} ({type(self)})")
            return converter(child).to_node()
        elif isinstance(child, Token):
            try:
                method = "convert_" + child.type.lower()
            except KeyError as e:
                raise InternalParserException(f"Unexpected token {e}")
            return getattr(self, method)(child)
        else:
            raise InternalParserException(f"Unexpected type {type(child)}")

    def to_node(self) -> Node:
        root = NoOpDecision(None, self.all_transitions)
        for child in self.tree.children:
            root.add_transition(self.convert(child))
        return root


class GroupConverter(TreeConverter):
    children = {
        'expression': None
    }

    def to_node(self) -> Node:
        expression = None
        repetition = None
        for child in self.tree.children:
            if isinstance(child, Tree):
                if child.data == 'expression':
                    expression = child
                elif child.data == 'repetition':
                    repetition = child
        if expression is None:
            raise InternalParserException("Expression not found")

        item = ExpressionConverter(expression).to_node()
        if repetition is None:
            return item
        else:
            root = NoOpDecision(None, self.all_transitions)
            rep = _parse_repetition(repetition)
            _repeat(root, item, rep)
            return root


class CharacterRangeConverter(TreeConverter):
    def to_node(self) -> Node:
        assert len(self.tree.children) == 2
        assert self.tree.data == 'character_range'
        first = self.tree.children[0]
        second = self.tree.children[1]
        assert isinstance(first, Token)
        assert isinstance(second, Token)
        if ord(first.value) > ord(second.value):
            raise RegexException(
                f"Invalid character range: {first.value}-{second.value}")
        root = NoOpDecision(None, self.all_transitions)
        root.add_transition(AppendCharsLeaf(None, True, first.value))
        root.add_transition(AppendCharsLeaf(None, True, second.value))
        # TODO: invalid samples
        # root.add_transition(AppendCharsLeaf(None, False, chr(ord(first.value)-1)))
        # root.add_transition(AppendCharsLeaf(None, False, chr(ord(second.value)+1)))
        return root


class CharacterGroupItemConverter(TreeConverter):
    children = {
        'character_range': CharacterRangeConverter
    }

    def convert_char(self, token: Token) -> Node:
        return AppendCharsLeaf(None, True, token.value)

    def convert_character_class(self, token: Token) -> Node:
        samples = {
            '\w': 'hello',
            '\W': '42',
            '\d': '42',
            '\D': 'world',
        }
        try:
            sample = samples[str(token)]
        except KeyError as e:
            raise RegexException(f"Unknown character class {e}")
        return AppendCharsLeaf(None, True, sample)


class CharacterGroupConverter(TreeConverter):

    children = {
        'character_group_item': CharacterGroupItemConverter
    }

    def to_node(self) -> Node:
        assert len(self.tree.children) >= 1
        invert = False
        start = 0
        if isinstance(self.tree.children[0], Token) and self.tree.children[0].type == 'CHARACTER_GROUP_NEGATIVE_MODIFIER':
            invert = True
            start = 1
        root = NoOpDecision(None, self.all_transitions)
        for child in self.tree.children[start:]:
            root.add_transition(self.convert(child))
        # TODO: use invert
        return root


class MatchCharacterClassConverter(TreeConverter):
    children = {
        'character_group': CharacterGroupConverter
    }


class EscapedCharacterConverter(TreeConverter):
    def convert_special_char(self, token: Token):
        return AppendCharsLeaf(None, True, str(token))


class MatchItemConverter(TreeConverter):
    children = {
        'match_character_class': MatchCharacterClassConverter,
        'escaped_character': EscapedCharacterConverter,
    }

    def convert_char(self, token: Token):
        return AppendCharsLeaf(None, True, str(token))

    def convert_match_any_character(self, token: Token):
        return NoOpLeaf(None, True)


class MatchConverter(TreeConverter):
    children = {
        'matchitem': MatchItemConverter,
        'match_character_class': MatchCharacterClassConverter,
    }
    all_transitions = False

    def to_node(self) -> Node:
        assert len(self.tree.children) in [1, 2]
        assert self.tree.children[0].data == 'matchitem'
        item = MatchItemConverter(self.tree.children[0]).to_node()
        if len(self.tree.children) == 2:
            repetition = self.tree.children[1]
            assert isinstance(repetition, Tree)
            assert repetition.data == 'repetition'
            rep = _parse_repetition(repetition)
            root = NoOpDecision(None, self.all_transitions)
            _repeat(root, item, rep)
            return root
        else:
            return item


class SubExpressionItemConverter(TreeConverter):
    children = {
        'match': MatchConverter,
        'group': GroupConverter,
    }

    def convert_anchor(self, token: Token):
        # TODO
        return NoOpLeaf(None, True)


class SubExpressionConverter(TreeConverter):
    children = {
        'subexpressionitem': SubExpressionItemConverter
    }
    all_transitions = True


class ExpressionConverter(TreeConverter):
    children = {
        'subexpression': SubExpressionConverter,
        'expression': None,
    }


class StartConverter(TreeConverter):
    children = {
        'expression': ExpressionConverter,
    }
    all_transitions = True

    def convert_start_of_string_anchor(self, token: Token):
        return AppendCharsLeaf(None, True, 'X')


GroupConverter.children['expression'] = ExpressionConverter
ExpressionConverter.children['expression'] = ExpressionConverter


def unescape(s: str) -> str:
    for i in ['\\', '{', '}', '.', '(', ')', '[', ']']:
        s = s.replace('\\' + i, i)
    return s


def parse(regex: str) -> Node:
    regex = unescape(regex)
    try:
        tree: Tree[Token] = _parser.parse(regex)
    except LarkError as e:
        raise RegexException(f"Cannot parse '{regex}' as regex: {e}")
    # print(tree.pretty())
    assert isinstance(tree, Tree)
    assert tree.data == 'start'
    root = StartConverter(tree).to_node()
    root.optimize()

    create_input = CreateInputNode()
    super_root = NoOpDecision(None, True)
    create_input.add_transition(super_root)
    super_root.add_transition(root)
    super_root.add_transition(FetchOutputNode())
    return create_input



from typing import Union, Generator, Tuple, List
from xml.etree.ElementTree import Element


def _get_tag(el: Element) -> str:
    _, _, tag_wo_namespace = el.tag.rpartition('}')
    return tag_wo_namespace


NormalizedXPathElement = Tuple[str, int]

class NormalizedXPath:
    """
    A simple xpath implementation, of the form
        /foo[0]/bar[0]/x[0]
    It always selects exactly one element.
    """

    def __init__(self, elements: List[NormalizedXPathElement] = None) -> None:
        self.elements = elements or []

    def __add__(self, other: NormalizedXPathElement) -> "NormalizedXPath":
        return NormalizedXPath(self.elements + [other])

    def __str__(self) -> str:
        return '/' + '/'.join([f"{i[0]}[{i[1]}]" for i in self.elements])

    def enumerate(self, element: Element) -> Generator[Tuple["NormalizedXPath", Element], None, None]:
        visited = {}
        for child in element:
            tag = _get_tag(child)
            if tag in visited:
                visited[tag] += 1
            else:
                visited[tag] = 0
            yield self + (tag, visited[tag]), child
